import warnings
import astropy.units as u

from cosipy.config import Configurator

from .modelmap import ModelMap
# import image deconvolution algorithms
from .RichardsonLucy import RichardsonLucy

class ImageDeconvolution:
    """
    A class to reconstruct all-sky images from COSI data based on image deconvolution methods.
    """

    def __init__(self):
        self._initial_model_map = None

    def set_data(self, data):
        """
        Set COSI dataset

        Parameters
        ----------
        data : :py:class:`cosipy.image_deconvolution.DataLoader`
            Data loader contaning an event histogram, a background model, a response matrix, and a coordsys_conversion_matrix.

        Notes
        -----
        cosipy.image_deconvolution.DataLoader may be removed in the future once the formats of event/background/response are fixed.
        In this case, this method will be also modified in the future.
        """

        self._data = data
        
        print("data for image deconvolution was set -> ", data)

    def read_parameterfile(self, parameter_filepath):
        """
        Read parameters from a yaml file.

        Parameters
        ----------
        parameter_filepath : str or pathlib.Path
            Path of parameter file.
        """

        self._parameter = Configurator.open(parameter_filepath)

        print("parameter file for image deconvolution was set -> ", parameter_filepath)

    @property
    def data(self):
        """
        Return the set data.
        """
        return self._data

    @property
    def parameter(self):
        """
        Return the set parameter.
        """
        return self._parameter

    def override_parameter(self, *args):
        """
        Override parameter

        Parameters
        ----------
        *args
            new parameter

        Examples
        --------
        >>> image_deconvolution.override_parameter("deconvolution:parameter_RL:iteration = 30")
        """
        self._parameter.override(args)

    @property
    def initial_model_map(self):
        """
        Return the initial model map.
        """
        if self._initial_model_map is None:
            warnings.warn("Need to initialize model map in the image_deconvolution instance")

        return self._initial_model_map

    def _check_model_response_consistency(self):
        """
        Check whether the axes of model map are consistent with those of the response matrix.

        Returns
        -------
        bool
            If True, their axes are consistent with each other.

        Notes
        -----
        It will be implemented in the future. Currently it always returns true.
        """
        #self._initial_model_map.axes["Ei"].axis_scale = self._data.image_response_dense_projected.axes["Ei"].axis_scale

        #return self._initial_model_map.axes["lb"] == self._data.image_response_dense_projected.axes["lb"] \
        #       and self._initial_model_map.axes["Ei"] == self._data.image_response_dense_projected.axes["Ei"]
        return True

    def initialize(self):
        """
        Initialize an image_deconvolution instance. It is mandatory to execute this method before running the image deconvolution.

        This method has three steps:
        1. generate a model map with properties (nside, energy bins, etc.) given in the parameter file.
        2. initialize a model map following an initial condition given in the parameter file
        3. load parameters for the image deconvolution
        """

        print("#### Initialization ####")

        ### check self._data ###
        ### this part will be removed in the future ###
        if self._data.response_on_memory == False:

            warnings.warn("In the image deconvolution, the option to not load the response on memory is currently not supported. Performing DataLoader.load_full_detector_response_on_memory().")
            self._data.load_full_detector_response_on_memory()

        if self._data.image_response_dense_projected is None:

            warnings.warn("The image_response_dense_projected has not been calculated. Performing DataLoader.calc_image_response_projected().")
            self._data.calc_image_response_projected()
        
        print("1. generating a model map") 
        parameter_model_property = Configurator(self._parameter['model_property'])
        self._initial_model_map = ModelMap(nside = parameter_model_property['nside'],
                                           energy_edges = parameter_model_property['energy_edges'] * u.keV, 
                                           scheme = parameter_model_property['scheme'], 
                                           coordsys = parameter_model_property['coordinate'])

        print("---- parameters ----")
        print(parameter_model_property.dump())
        
        print("2. initializing the model map ...")
        parameter_model_initialization = Configurator(self._parameter['model_initialization'])

        algorithm_name = parameter_model_initialization['algorithm']

        self._initial_model_map.set_values_from_parameters(algorithm_name, 
                                                           parameter_model_initialization['parameter_'+algorithm_name])

        if not self._check_model_response_consistency():
            return

        print("---- parameters ----")
        print(parameter_model_initialization.dump())

        print("3. registering the deconvolution algorithm ...")
        parameter_deconvolution = Configurator(self._parameter['deconvolution'])
        self._deconvolution = self.register_deconvolution_algorithm(parameter_deconvolution)

        print("---- parameters ----")
        print(parameter_deconvolution.dump())

        print("#### Done ####")
        print("")

    def register_deconvolution_algorithm(self, parameter_deconvolution):
        """
        Register parameters for image deconvolution on a deconvolution instance.

        Parameters
        ----------
        parameter_deconvolution : :py:class:`cosipy.config.Configurator`
            Parameters for the image deconvolution methods.

        Notes
        -----
        Currently only RichardsonLucy algorithm is implemented.

        ***An example of parameters for RL algorithm***
        algorithm: "RL"
        parameter_RL:
            iteration: 10 
            # number of iterations
            acceleration: True 
            # whether the accelerated ML-EM algorithm (Knoedlseder+99) is used
            alpha_max: 10.0 
            # the maximum value for the acceleration alpha parameter
            save_results_each_iteration: False 
            # whether a updated model map, detal map, likelihood etc. are save at the end of each iteration
            response_weighting: True 
            # whether a factor $w_j = (\sum_{i} R_{ij})^{\beta}$ for weighting the delta image is introduced 
            # see Knoedlseder+05, Siegert+20
            response_weighting_index: 0.5 
            # $\beta$ in the above equation
            smoothing: True 
            # whether a Gaussian filter is used (see Knoedlseder+05, Siegert+20)
            smoothing_FWHM: 2.0 #deg 
            # the FWHM of the Gaussian in the filter 
            background_normalization_fitting: False 
            # whether the background normalization is optimized at each iteration. 
            # As for now, the same single background normalization factor is used in all of the time bins
            background_normalization_range: [0.01, 10.0]
            # the range of the normalization factor. it should be positive.
        """

        algorithm_name = parameter_deconvolution['algorithm']

        if algorithm_name == 'RL':
            parameter_RL = Configurator(parameter_deconvolution['parameter_RL'])
            _deconvolution = RichardsonLucy(self._initial_model_map, self._data, parameter_RL)
#        elif algorithm_name == 'MaxEnt':
#            parameter = self.parameter['deconvolution']['parameter_MaxEnt']
#            self.deconvolution == ...

        return _deconvolution

    def run_deconvolution(self):
        """
        Perform the image deconvolution. Make sure that the initialize method has been conducted.
        
        Returns
        -------
        list
            List containing results (reconstructed image, likelihood etc) at each iteration. 
        """
        print("#### Deconvolution Starts ####")
        
        all_result = []
        for result in self._deconvolution.iteration():
            all_result.append(result)
            ### can perform intermediate check ###
            #...
            ###

        print("#### Done ####")
        print("")
        return all_result

#    def analyze_result(self):
#        pass



import logging
logger = logging.getLogger(__name__)

import numpy as np

import yaml

from pathlib import Path

class Configurator:
    """
    Access to configuration parameters. 

    The parameters are organized in nested dictionarys. They set values
    can be a number, a string or a list. They can be accessed as e.g. 
    :code:`config['group:subgroup:subsubgroup:parameter']`
    
    Parameters
    -----
    config_path : Path
        Path to yaml configuration file.
    """

    def __init__(self, config = None):

        if config is None:
            self._config = None
        else:
            self._config = config

    @classmethod
    def open(cls, config_path):

        new = cls()
        
        new._config = {}

        new.config_path = Path(config_path)
        
        logger.info("Using configuration file at %s", config_path)

        with open(new.config_path) as f:
            new._config = yaml.safe_load(f)

        return new
            
    def __getitem__(self, key):

        value = self._config

        keys = key.split(':')
        
        for n,key in enumerate(keys):

            try:
                value = value[key]
            except KeyError:
                raise KeyError("Parameter '{}' doesn't exists".\
                               format(":".join(keys[:n+1])))
                
        return value
        
    def __setitem__(self, key, value):

        keys = key.split(':')
        
        elem = self._config

        for n,key in enumerate(keys[:-1]):

            if key not in elem:
                raise KeyError("Parameter '{}' doesn't exists".\
                               format(":".join(keys[:n+1])))
            
            elem = elem[key]

        if keys[-1] not in elem:
            raise KeyError("Parameter '{}' doesn't exists".\
                           format(":".join(keys)))
            
        elem[keys[-1]] = value

    def get(self, key, default = None):
        """
        Returns the value for key in the config. If not found returns a default value.

        Parameters
        ----------
        key :  str
            Key

        default
            default value
        """

        try:
            value = self[key]
        except KeyError:
            value = default
 
        return value
        
    def override(self, *args):
        """
        Override one or more parameter by parsings strings of the form
        :code:`'group:subgroup:parameter = value'`. Value is parsed the same way
        as yaml would do it, so if it is a string use quotes. 

        The foreseen use case scenario is to let the uses override parameters 
        from command line e.g. :code:`--override "group:parameter = new_value"`.
        Note that the values can be override directly in code by using
        :code:`config['group:parameter'] = new_value"`.

        Parameters
        ----------
        args : string, array
            String(s) to parse and override parameters. 
        """
        
        if len(args) == 1:
        
            if np.isscalar(args[0]):
                # Standard, single key
                key,value = args[0].split("=")

                self[key.strip()] = yaml.safe_load(value.strip())

            else:
                #Recursive
                for arg in args[0]:
                    self.override(arg)

        else:
            #Recursive
            for arg in args:
                self.override(arg)
            

    def absolute_path(self, path):
        """
        Turn a path relative to the location of the config file absolute

        Parameters
        ----------
        path: Path
            Relative path

        Returns
        -------
        Path
        """

        path = Path(path).expanduser()

        if path.is_absolute():
            return path
        else:
            return (self.config_path.parent / path).resolve()

    def dump(self, *args, **kwargs):
        """
        Dump the configuration contents to a file or a string.

        Parameters
        ----------
        args, kwargs
            All arguments are passed to yaml.dump(). By default it returns as 
            string. You can specify a file or stream to dump it into in the
            first argument.
        """
        
        return yaml.dump(self._config, *args, **kwargs)
        



import warnings
import astropy.units as u
import numpy as np

from histpy import Histogram, Axes, Axis, HealpixAxis

from cosipy.threeml.custom_functions import get_integrated_spectral_model

class ModelMap(Histogram):
    """
    Photon flux maps in given energy bands. 2-dimensional histogram.

    Attributes
    ----------
    nside : int
        Healpix NSIDE parameter.
    energy_edges : :py:class:`np.array`
        Bin edges for energies. We recommend to use a Quantity array with the unit of keV.
    scheme : str, default 'ring'
        Healpix scheme. Either 'ring', 'nested'.
    coordsys : str or :py:class:`astropy.coordinates.BaseRepresentation`, default is 'galactic'
        Instrinsic coordinates of the map. The default is 'galactic'.
    label_image : str, default 'lb'
        The label name of the healpix axis.
    label_energy : str, default 'Ei'
        The label name of the energy axis. The default is 'Ei'.
    """

    def __init__(self,
                 nside,
                 energy_edges,
                 scheme = 'ring',
                 coordsys = 'galactic',
                 label_image = 'lb',
                 label_energy = 'Ei'
                 ):

        if energy_edges.unit != u.keV:

            warnings.warn(f"The unit of the given energy_edges is {energy_edges.unit}. It is converted to keV.")
            energy_edges = energy_edges.to('keV')

        self.image_axis = HealpixAxis(nside = nside,
                                      scheme = scheme,
                                      coordsys = coordsys,
                                      label = label_image)

        self.energy_axis = Axis(edges = energy_edges, label = label_energy, scale = "log")

        axes = Axes([self.image_axis, self.energy_axis])

        super().__init__(axes, sparse = False, unit = 1 / u.s / u.cm**2 / u.sr) # unit might be specified in the input parameter.

    def set_values_from_parameters(self, algorithm_name, parameter):
        """
        Set the values of this model map accordinng to the specified algorithm. 

        Parameters
        ----------
        algorithm_name : str
            Algorithm name to fill the values.  
        parameter : py:class:`cosipy.config.Configurator`
            Parameters for the specified algorithm.

        Notes
        -----
        Currently algorithm_name can be only 'flat'. All of the pixel values in each energy bins will set to the given value.
        parameter should be {'values': [ flux value at 1st energy bin (without unit), flux value at 2nd energy bin, ...]}.
        """

        if algorithm_name == "flat":
            for idx, value in enumerate(parameter['values']):
                self[:,idx:idx+1] = value * self.unit
    #    elif algorithm_name == ... 
    #       ...

    def set_values_from_extendedmodel(self, extendedmodel):
        """
        Set the values of this model map accordinng to the given astromodels.ExtendedSource.

        Parameters
        ----------
        extendedmodel : :py:class:`astromodels.ExtendedSource`
            Extended source model
        """

        integrated_flux = get_integrated_spectral_model(spectrum = extendedmodel.spectrum.main.shape,
                                                        eaxis = self.energy_axis)
        
        npix = self.image_axis.npix
        coords = self.image_axis.pix2skycoord(np.arange(npix))

        normalized_map = extendedmodel.spatial_shape(coords.l.deg, coords.b.deg) / u.sr

        self[:] = np.tensordot(normalized_map, integrated_flux.contents, axes = 0) 



import copy
import numpy as np
from tqdm.autonotebook import tqdm

from histpy import Histogram

from .deconvolution_algorithm_base import DeconvolutionAlgorithmBase

class RichardsonLucy(DeconvolutionAlgorithmBase):
    """
    A class for the RichardsonLucy algorithm. 
    The algorithm here is based on Knoedlseder+99, Knoedlseder+05, Siegert+20.
    """

    def __init__(self, initial_model_map, data, parameter):
        DeconvolutionAlgorithmBase.__init__(self, initial_model_map, data, parameter)

        self.loglikelihood = None

        self.alpha_max = parameter.get('alpha_max', 1.0)

        self.do_response_weighting = parameter.get('response_weighting', False)

        self.do_smoothing = parameter.get('smoothing', False)

        self.do_bkg_norm_fitting = parameter.get('background_normalization_fitting', False)

        if self.do_bkg_norm_fitting:
            self.bkg_norm_range = parameter.get('background_normalization_range', [0.5, 1.5])

        print("... calculating the expected events with the initial model map ...")
        self.expectation = self.calc_expectation(self.initial_model_map, self.data)

        if self.do_response_weighting:
            print("... calculating the response weighting filter...")

            response_weighting_index = parameter.get('response_weighting_index', 0.5)

            self.response_weighting_filter = data.image_response_dense_projected.contents / np.max(data.image_response_dense_projected.contents) 

            self.response_weighting_filter = self.response_weighting_filter**response_weighting_index

        if self.do_smoothing:
            print("... calculating the gaussian filter...")

            self.smoothing_sigma = parameter['smoothing_FWHM'] / 2.354820 # degree

            self.gaussian_filter = self.calc_gaussian_filter(self.smoothing_sigma)

    def pre_processing(self):
        pass

    def Estep(self):
        """
        Notes
        -----
        Expect count histogram is calculated in the post processing.
        """
        print("... skip E-step ...")

    def Mstep(self):
        """
        M-step in RL algorithm.

        Notes
        -----
        Background normalization is also optimized based on a generalized RL algirithm.
        Currenly we use a signle normalization parameter. 
        In the future, the normalization will be optimized for each background group defined in some file.
        """
        # Currenly (2024-01-12) this method can work for both local coordinate CDS and in galactic coordinate CDS.
        # This is just because in DC2 the rotate response for galactic coordinate CDS does not have an axis for time/scatt binning.
        # However it is likely that it will have such an axis in the future in order to consider background variability depending on time and pointign direction etc.
        # Then, the implementation here will not work. Thus, keep in mind that we need to modify it once the response format is fixed.

        diff = self.data.event_dense / self.expectation - 1

        delta_map_part1 = self.model_map / self.data.image_response_dense_projected
        delta_map_part2 = Histogram(self.model_map.axes, unit = self.data.image_response_dense_projected.unit)

        if self.data.response_on_memory == True:
            diff_x_response = np.tensordot(diff.contents, self.data.image_response_dense.contents, axes = ([1,2,3], [2,3,4])) 
            # [Time/ScAtt, Em, Phi, PsiChi] x [NuLambda, Ei, Em, Phi, PsiChi] -> [Time/ScAtt, NuLambda, Ei]

            delta_map_part2[:] = np.tensordot(self.data.coordsys_conv_matrix.contents, diff_x_response, axes = ([0,2], [0,1])) \
                                 * diff_x_response.unit * self.data.coordsys_conv_matrix.unit
            # [Time/ScAtt, lb, NuLambda] x [Time/ScAtt, NuLambda, Ei] -> [lb, Ei]
            # note that coordsys_conv_matrix is the sparse, so the unit should be recovered.

        else:
            for ipix in tqdm(range(self.npix_local)):
                if self.data.is_miniDC2_format == True:
                    response_this_pix = np.sum(self.data.full_detector_response[ipix].to_dense(), axis = (4,5)) # [Ei, Em, Phi, PsiChi]
                else:
                    response_this_pix = self.data.full_detector_response[ipix].to_dense() # [Ei, Em, Phi, PsiChi]

                diff_x_response_this_pix = np.tensordot(diff.contents, response_this_pix, axes = ([1,2,3], [1,2,3]))
                # [Time/ScAtt, Em, Phi, PsiChi] x [Ei, Em, Phi, PsiChi] -> [Time/ScAtt, Ei]

                delta_map_part2 += np.tensordot(self.data.coordsys_conv_matrix[:,:,ipix], diff_x_response_this_pix, axes = ([0],[0])) \
                                   * diff_x_response_this_pix.unit * self.data.coordsys_conv_matrix.unit #lb, Ei
                # [Time/ScAtt, lb] x [Time/ScAtt, Ei] -> [lb, Ei]

        self.delta_map = delta_map_part1 * delta_map_part2

        if self.do_bkg_norm_fitting:
            self.bkg_norm += self.bkg_norm * np.sum(diff * self.data.bkg_dense) / np.sum(self.data.bkg_dense)

            if self.bkg_norm < self.bkg_norm_range[0]:
                self.bkg_norm = self.bkg_norm_range[0]
            elif self.bkg_norm > self.bkg_norm_range[1]:
                self.bkg_norm = self.bkg_norm_range[1]

    def post_processing(self):
        """
        Here three processes will be performed.
        - response weighting filter: the delta map is renormalized as pixels with large exposure times will have more feedback.
        - gaussian smoothing filter: the delta map is blurred with a Gaussian function.
        - acceleration of RL algirithm: the normalization of delta map is increased as long as the updated image has no non-negative components.
        """

        if self.do_response_weighting:
            self.delta_map[:,:] *= self.response_weighting_filter

        if self.do_smoothing:
            self.delta_map[:,:] = np.tensordot(self.gaussian_filter.contents, self.delta_map.contents, axes = [[0], [0]])

        self.alpha = self.calc_alpha(self.delta_map, self.model_map)

        self.processed_delta_map = self.delta_map * self.alpha

        self.model_map += self.processed_delta_map 

        print("... calculating the expected events with the updated model map ...")
        self.expectation = self.calc_expectation(self.model_map, self.data)

    def check_stopping_criteria(self, i_iteration):
        """
        If i_iteration is smaller than iteration_max, the iterative process will continue.

        Returns
        -------
        bool
        """
        if i_iteration < self.iteration_max:
            return False
        return True

    def register_result(self, i_iteration):
        """
        The values below are stored at the end of each iteration.
        - iteration: iteration number
        - model_map: updated image
        - delta_map: delta map after M-step 
        - processed_delta_map: delta map after post-processing
        - alpha: acceleration parameter in RL algirithm
        - background_normalization: optimized background normalization
        - loglikelihood: log-likelihood
        """
        loglikelihood = self.calc_loglikelihood(self.data, self.model_map, self.expectation)

        this_result = {"iteration": i_iteration, 
                       "model_map": copy.deepcopy(self.model_map), 
                       "delta_map": copy.deepcopy(self.delta_map),
                       "processed_delta_map": copy.copy(self.processed_delta_map),
                       "alpha": self.alpha, 
                       "background_normalization": self.bkg_norm,
                       "loglikelihood": loglikelihood}

        self.result = this_result

    def save_result(self, i_iteration):
        self.result["model_map"].write(f"model_map_itr{i_iteration}.hdf5", overwrite = True)
        self.result["delta_map"].write(f"delta_map_itr{i_iteration}.hdf5", overwrite = True)
        self.result["processed_delta_map"].write(f"processed_delta_map_itr{i_iteration}.hdf5", overwrite = True)

        with open(f"result_itr{i_iteration}.dat", "w") as f:
            f.write(f'alpha: {self.result["alpha"]}\n')
            f.write(f'loglikelihood: {self.result["loglikelihood"]}\n')
            f.write(f'background_normalization: {self.result["background_normalization"]}\n')

    def show_result(self, i_iteration):
        print(f'    alpha: {self.result["alpha"]}')
        print(f'    loglikelihood: {self.result["loglikelihood"]}')
        print(f'    background_normalization: {self.result["background_normalization"]}')

    def calc_alpha(self, delta, model_map, almost_zero = 1e-5): #almost_zero is needed to prevent producing a flux below zero
        """
        Calculate the acceleration parameter in RL algorithm.

        Returns
        -------
        float
            Acceleration parameter
        """
        alpha = -1.0 / np.min( delta / model_map ) * (1 - almost_zero)
        alpha = min(alpha, self.alpha_max)
        if alpha < 1.0:
            alpha = 1.0
        return alpha



import gc
from tqdm.autonotebook import tqdm
import numpy as np
import healpy as hp
import astropy.units as u
from astropy.coordinates import angular_separation

from histpy import Histogram, Axes, Axis

class DeconvolutionAlgorithmBase(object):
    """
    A base class for image deconvolution algorithms.
    A subclass should override these methods:

    - pre_processing
    - Estep
    - Mstep
    - post_processing
    - check_stopping_criteria
    - register_result
    - save_result
    - show_result
    
    When the method run_deconvolution is called in ImageDeconvolution class, 
    the iteration method in this class is called for each iteration.

    Attributes
    ----------
    initial_model_map : :py:class:`cosipy.image_deconvolution.ModelMap`
        Initial values for reconstructed images
    data : :py:class:`cosipy.image_deconvolution.DataLoader`
        COSI data set 
    parameter : py:class:`cosipy.config.Configurator`
        Parameters for a deconvolution algorithm
    """

    def __init__(self, initial_model_map, data, parameter):

        self.data = data 

        self.parameter = parameter 

        self.initial_model_map = initial_model_map

        # image axis (model space)
        image_axis = initial_model_map.axes['lb']

        self.nside_model = image_axis.nside
        self.npix_model = image_axis.npix
        self.pixelarea_model = hp.nside2pixarea(self.nside_model) * u.sr

        # energy axis (model space)
        energy_axis = initial_model_map.axes['Ei']
        self.n_energyband_model = energy_axis.nbins

        # NuLambda axis (Compton data space)
        response_axis = data.full_detector_response.axes['NuLambda']

        self.nside_local = response_axis.nside
        self.npix_local = response_axis.npix
        self.pixelarea_local = hp.nside2pixarea(self.nside_local) * u.sr

        # reconstructed image and related data
        self.model_map = None
        self.delta_map = None
        self.processed_delta_map = None
        self.bkg_norm = 1.0

        self.result = None

        self.expectation = None
        
        # parameters of the iteration
        self.iteration_max = parameter['iteration']

        self.save_result = parameter.get("save_results_each_iteration", False)

    def pre_processing(self):
        """
        pre-processing for each iteration
        """
        pass

    def Estep(self):
        """
        E-step. Basically expected counts are calculated here (or at the end of iteration in some cases).
        """
        pass

    def Mstep(self):
        """
        M-step. Basically a first feedback to a model map (delta map) is calculated here.
        """
        pass

    def post_processing(self):
        """
        Post-processing. For example, filters like gaussian smoothing are applied to the delta map in this step.
        """
        pass

    def check_stopping_criteria(self, i_iteration):
        """
        Check whether iteration process should be continued or stopped.
        """
        if i_iteration < 0:
            return False
        return True

    def register_result(self, i_iteration):
        """
        Register results at the end of each iteration. Users can define what kinds of values will be stored in this method.
        """
        this_result = {"iteration": self.i_iteration + 1}
        self.result = this_result

    def save_result(self, i_iteration):
        """
        Save some results at the end of each iteration.
        """
        pass

    def show_result(self, i_iteration):
        """
        Show some results at the end of each iteration. 
        """
        pass

    def iteration(self):
        """
        Perform one iteration of image deconvolution.
        This method should not be overrided in subclasses.
        """
        self.model_map = self.initial_model_map

        stop_iteration = False
        for i_iteration in tqdm(range(1, self.iteration_max + 1)):
            gc.collect()

            if stop_iteration:
                break

            print("  Iteration {}/{} ".format(i_iteration, self.iteration_max))

            print("--> pre-processing")
            self.pre_processing()

            print("--> E-step")
            self.Estep()
            gc.collect()

            print("--> M-step")
            self.Mstep()
            gc.collect()
            
            print("--> post-processing")
            self.post_processing()

            print("--> checking stopping criteria")
            stop_iteration = self.check_stopping_criteria(i_iteration)
            print("--> --> {}".format("stop" if stop_iteration else "continue"))

            print("--> registering results")
            self.register_result(i_iteration)

            print("--> showing results")
            self.show_result(i_iteration)
            
            if self.save_result == True:
                print("--> saving results")
                self.save_result(i_iteration)

            gc.collect()

            yield self.result
    
    def calc_expectation(self, model_map, data, almost_zero = 1e-12):
        """
        Calculate expected counts from a given model map.

        Parameters
        ----------
        model_map : :py:class:`cosipy.image_deconvolution.ModelMap`
            Model map
        data : :py:class:`cosipy.image_deconvolution.DataLoader`
            COSI data set 
        almost_zero : float, default 1e-12
            In order to avoid zero components in extended count histogram, a tiny offset is introduced.
            It should be small enough not to effect statistics.

        Returns
        -------
        :py:class:`histpy.Histogram`
            Expected count histogram

        Notes
        -----
        This method should be implemented in a more general class, for example, extended source response class in the future.
        """
        # Currenly (2024-01-12) this method can work for both local coordinate CDS and in galactic coordinate CDS.
        # This is just because in DC2 the rotate response for galactic coordinate CDS does not have an axis for time/scatt binning.
        # However it is likely that it will have such an axis in the future in order to consider background variability depending on time and pointign direction etc.
        # Then, the implementation here will not work. Thus, keep in mind that we need to modify it once the response format is fixed.

        expectation = Histogram(data.event_dense.axes) 

        map_rotated = np.tensordot(data.coordsys_conv_matrix.contents, model_map.contents, axes = ([1], [0])) 
        # ['Time/ScAtt', 'lb', 'NuLambda'] x ['lb', 'Ei'] -> [Time/ScAtt, NuLambda, Ei]
        map_rotated *= data.coordsys_conv_matrix.unit * model_map.unit
        map_rotated *= self.pixelarea_model
        # data.coordsys_conv_matrix.contents is sparse, so the unit should be restored.
        # the unit of map_rotated is 1/cm2 ( = s * 1/cm2/s/sr * sr)

        if data.response_on_memory == True:
            expectation[:] = np.tensordot( map_rotated, data.image_response_dense.contents, axes = ([1,2], [0,1]))
        else:
            for ipix in tqdm(range(self.npix_local)):
                response_this_pix = np.sum(data.full_detector_response[ipix].to_dense(), axis = (4,5)) # ['Ei', 'Em', 'Phi', 'PsiChi', 'SigmaTau', 'Dist'] -> ['Ei', 'Em', 'Phi', 'PsiChi']
                expectation += np.tensordot(map_rotated[:,ipix,:], response_this_pix, axes = ([1], [0]))

        expectation += data.bkg_dense * self.bkg_norm
        expectation += almost_zero
        
        return expectation

    def calc_loglikelihood(self, data, model_map, expectation = None):
        """
        Calculate log-likelihood from given data and model/expectation.

        Parameters
        ----------
        data : :py:class:`cosipy.image_deconvolution.DataLoader`
            COSI data set 
        model_map : :py:class:`cosipy.image_deconvolution.ModelMap`
            Model map
        expectation : :py:class:`histpy.Histogram` or None, default None
            Expected count histogram.
            If it is not given, expected counts will be calculated in this method.

        Returns
        -------
        float
            Log-likelood

        Notes
        -----
        The parameter expectation may be a mandatory parameter in the future.
        """
        if expectation is None:
            expectation = self.calc_expectation(model_map, data)

        loglikelood = np.sum( data.event_dense * np.log(expectation) ) - np.sum(expectation)

        return loglikelood

    def calc_gaussian_filter(self, sigma):
        """
        Calculate a gaussian filter for post-processing.

        Parameters
        ----------
        sigma: float
            Sigma for Gaussian function. It should be in degrees, but unitless.

        Returns
        -------
        :py:class:`histpy.Histogram`
            Gaussian filter. 2-dimensional histogram.
        """

        gaussian_filter = Histogram( Axes( [Axis(edges = np.arange(self.npix_model+1)), Axis(edges = np.arange(self.npix_model+1))] ), sparse = False)

        for ipix in tqdm(range(self.npix_model)):

            lon_ref, lat_ref = hp.pix2ang(self.nside_model, ipix, nest = False, lonlat = True)

            lon, lat = hp.pix2ang(self.nside_model, np.arange(self.npix_model), nest = False, lonlat = True)

            delta_ang = angular_separation(lon_ref * u.deg, lat_ref * u.deg, lon * u.deg, lat * u.deg).to('deg').value
    
            gaussian_filter[ipix,:] = np.exp( - 0.5 * delta_ang**2 / sigma**2)
    
            gaussian_filter[ipix,:] /= np.sum(gaussian_filter[ipix,:])  

        return gaussian_filter



from histpy import Histogram, Axes, Axis

import astropy.units as u

from astropy.units import Quantity

from scipy import integrate

from threeML import DiracDelta, Constant, Line, Quadratic, Cubic, Quartic, StepFunction, StepFunctionUpper, Cosine_Prior, Uniform_prior, PhAbs, Gaussian


class PointSourceResponse(Histogram):
    """
    Handles the multi-dimensional matrix that describes the expected
    response of the instrument for a particular point in the sky.

    Parameters
    ----------
    axes : :py:class:`histpy.Axes`
        Binning information for each variable. The following labels are expected:\n
        - ``Ei``: Real energy
        - ``Em``: Measured energy. Optional
        - ``Phi``: Compton angle. Optional.
        - ``PsiChi``:  Location in the Compton Data Space (HEALPix pixel). Optional.
        - ``SigmaTau``: Electron recoil angle (HEALPix pixel). Optional.
        - ``Dist``: Distance from first interaction. Optional.
    contents : array, :py:class:`astropy.units.Quantity` or :py:class:`sparse.SparseArray`
        Array containing the differential effective area convolved with wht source exposure.
    unit : :py:class:`astropy.units.Unit`, optional
        Physical units, if not specified as part of ``contents``. Units of ``area*time``
        are expected.
    """
    
    @property
    def photon_energy_axis(self):
        """
        Real energy bins (``Ei``).

        Returns
        -------
        :py:class:`histpy.Axes`
        """
        
        return self.axes['Ei']
       
    def get_expectation(self, spectrum):
        """
        Convolve the response with a spectral hypothesis to obtain the expected
        excess counts from the source.

        Parameters
        ----------
        spectrum : :py:class:`threeML.Model`
            Spectral hypothesis.

        Returns
        -------
        :py:class:`histpy.Histogram`
             Histogram with the expected counts on each analysis bin
        """
        
        eaxis = self.photon_energy_axis
        
        spectrum_unit = None

        for item in spectrum.parameters:
            if getattr(spectrum, item).is_normalization == True:
                spectrum_unit = getattr(spectrum, item).unit
                break
                
        if spectrum_unit == None:
            if isinstance(spectrum, Constant):
                spectrum_unit = spectrum.k.unit
            elif isinstance(spectrum, Line) or isinstance(spectrum, Quadratic) or isinstance(spectrum, Cubic) or isinstance(spectrum, Quartic):
                spectrum_unit = spectrum.a.unit
            elif isinstance(spectrum, StepFunction) or isinstance(spectrum, StepFunctionUpper) or isinstance(spectrum, Cosine_Prior) or isinstance(spectrum, Uniform_prior) or isinstance(spectrum, DiracDelta): 
                spectrum_unit = spectrum.value.unit
            elif isinstance(spectrum, PhAbs):
                spectrum_unit = u.dimensionless_unscaled
            elif isinstance(spectrum, Gaussian):
                spectrum_unit = spectrum.F.unit / spectrum.sigma.unit 
            else:
                try:
                    spectrum_unit = spectrum.K.unit
                except:
                    raise RuntimeError("Spectrum not yet supported because units of spectrum are unknown.")
                    
        if isinstance(spectrum, DiracDelta):
            flux = Quantity([spectrum.value.value * spectrum_unit * lo_lim.unit if spectrum.zero_point.value >= lo_lim/lo_lim.unit and spectrum.zero_point.value <= hi_lim/hi_lim.unit else 0 * spectrum_unit * lo_lim.unit
                             for lo_lim,hi_lim
                             in zip(eaxis.lower_bounds, eaxis.upper_bounds)])
        else:
            flux = Quantity([integrate.quad(spectrum, lo_lim/lo_lim.unit, hi_lim/hi_lim.unit)[0] * spectrum_unit * lo_lim.unit
                             for lo_lim,hi_lim
                             in zip(eaxis.lower_bounds, eaxis.upper_bounds)])
        
        flux = self.expand_dims(flux.value, 'Ei') * flux.unit

        expectation = self * flux
        
        return expectation



import warnings
import numpy as np
from tqdm.autonotebook import tqdm
import astropy.units as u

from histpy import Histogram, Axes

from cosipy.response import FullDetectorResponse
from cosipy.data_io import BinnedData
from .coordsys_conversion_matrix import CoordsysConversionMatrix

class DataLoader(object):
    """
    A class to manage data for image analysis, 
    namely event data, background model, response, coordsys conversion matrix.
    Ideally, these data should be input directly to ImageDeconvolution class,
    but considering their data formats are not fixed, this class is introduced.
    The purpose of this class is to check the consistency between input data and calculate intermediate files etc.
    In the future, this class may be removed or hidden in ImageDeconvolution class.
    """

    def __init__(self):
        self.event_dense = None
        self.bkg_dense = None
        self.full_detector_response = None
        self.coordsys_conv_matrix = None

        self.is_miniDC2_format = False

        self.response_on_memory = False

        self.image_response_dense_projected = None

    @classmethod
    def load(cls, event_binned_data, bkg_binned_data, rsp, coordsys_conv_matrix, is_miniDC2_format = False):
        """
        Load data

        Parameters
        ----------
        event_binned_data : :py:class:`histpy.Histogram`
            Event histogram
        bkg_binned_data : :py:class:`histpy.Histogram`
            Background model
        rsp : :py:class:`histpy.Histogram` or :py:class:`cosipy.response.FullDetectorResponse`
            Response
        coordsys_conv_matrix : :py:class:`cosipy.image_deconvolution.CoordsysConversionMatrix`
            Coordsys conversion matrix 
        is_miniDC2_format : bool, default False
            Whether the file format is for mini-DC2. It will be removed in the future.

        Returns
        -------
        :py:class:`cosipy.image_deconvolution.DataLoader`
            DataLoader instance containing the input data set
        """

        new = cls()

        new.event_dense = event_binned_data.to_dense()

        new.bkg_dense = bkg_binned_data.to_dense()

        new.full_detector_response = rsp

        new.coordsys_conv_matrix = coordsys_conv_matrix

        new.is_miniDC2_format = is_miniDC2_format

        return new
    
    @classmethod
    def load_from_filepath(cls, event_hdf5_filepath = None, event_yaml_filepath = None, 
                           bkg_hdf5_filepath = None, bkg_yaml_filepath = None, 
                           rsp_filepath = None, ccm_filepath = None,
                           is_miniDC2_format = False):
        """
        Load data from file pathes

        Parameters
        ----------
        event_hdf5_filepath : str or None, default None
            File path of HDF5 file for event histogram.
        event_yaml_filepath : str or None, default None
            File path of yaml file to read the HDF5 file.
        bkg_hdf5_filepath : str or None, default None
            File path of HDF5 file for background model.
        bkg_yaml_filepath : str or None, default None
            File path of yaml file to read the HDF5 file.
        rsp_filepath : str or None, default None
            File path of the response matrix.
        ccm_filepath : str or None, default None
            File path of the coordsys conversion matrix.
        is_miniDC2_format : bool, default False
            Whether the file format is for mini-DC2. should be removed in the future.

        Returns
        -------
        :py:class:`cosipy.image_deconvolution.DataLoader`
            DataLoader instance containing the input data set
        """

        new = cls()

        new.set_event_from_filepath(event_hdf5_filepath, event_yaml_filepath)

        new.set_bkg_from_filepath(bkg_hdf5_filepath, bkg_yaml_filepath)
        
        new.set_rsp_from_filepath(rsp_filepath)

        new.set_ccm_from_filepath(ccm_filepath)

        new.is_miniDC2_format = is_miniDC2_format

        return new

    def set_event_from_filepath(self, hdf5_filepath, yaml_filepath):
        """
        Load event data from file pathes

        Parameters
        ----------
        hdf5_filepath : str
            File path of HDF5 file for event histogram.
        yaml_filepath : str
            File path of yaml file to read the HDF5 file.
        """

        self._event_hdf5_filepath = hdf5_filepath
        self._event_yaml_filepath = yaml_filepath

        print(f'... loading event from {hdf5_filepath} and {yaml_filepath}')

        event = BinnedData(self._event_yaml_filepath)
        event.load_binned_data_from_hdf5(self._event_hdf5_filepath)

        self.event_dense = event.binned_data.to_dense()

        print("... Done ...")

    def set_bkg_from_filepath(self, hdf5_filepath, yaml_filepath):
        """
        Load background model from file pathes

        Parameters
        ----------
        hdf5_filepath : str
            File path of HDF5 file for background model.
        yaml_filepath : str
            File path of yaml file to read the HDF5 file.
        """

        self._bkg_hdf5_filepath = hdf5_filepath
        self._bkg_yaml_filepath = yaml_filepath

        print(f'... loading background from {hdf5_filepath} and {yaml_filepath}')

        bkg = BinnedData(self._bkg_yaml_filepath)
        bkg.load_binned_data_from_hdf5(self._bkg_hdf5_filepath)

        self.bkg_dense = bkg.binned_data.to_dense()

        print("... Done ...")

    def set_rsp_from_filepath(self, filepath):
        """
        Load response matrix from file pathes

        Parameters
        ----------
        filepath : str
            File path of the response matrix.
        """

        self._rsp_filepath = filepath

        print(f'... loading full detector response from {filepath}')

        self.full_detector_response = FullDetectorResponse.open(self._rsp_filepath) 

        print("... Done ...")

    def set_ccm_from_filepath(self, filepath):
        """
        Load coordsys conversion matrix from file pathes

        Parameters
        ----------
        filepath : str
            File path of the coordsys conversion matrix.
        """

        self._ccm_filepath = filepath
        
        print(f'... loading coordsys conversion matrix from {filepath}')

        self.coordsys_conv_matrix = CoordsysConversionMatrix.open(self._ccm_filepath)

        print("... Done ...")

    def _check_file_registration(self):
        """
        Check whether files are loaded.

        Returns
        -------
        bool
            True if all required files are loaded.
        """

        print(f"... checking the file registration ...")

        if self.event_dense and self.bkg_dense \
        and self.full_detector_response and self.coordsys_conv_matrix:

            print(f"    --> pass")
            return True
        
        return False

    def _check_axis_consistency(self):
        """
        Check whether the axes of event/background/response are consistent with each other.

        Returns
        -------
        bool
            True if their axes are consistent.
        """
        
        print(f"... checking the axis consistency ...")

        # check the axes of the event/background files
        if self.coordsys_conv_matrix.binning_method == 'Time':
            axis_name = ['Time', 'Em', 'Phi', 'PsiChi']

        elif self.coordsys_conv_matrix.binning_method == 'ScAtt':
            axis_name = ['ScAtt', 'Em', 'Phi', 'PsiChi']

        for name in axis_name:
            if not self.event_dense.axes[name] == self.bkg_dense.axes[name]:
                print(f"Warning: the axis {name} is not consistent between the event and background!")
                return False

        # check the axes of the event/response files
        axis_name = ['Em', 'Phi', 'PsiChi']
        
        for name in axis_name:
            if not self.event_dense.axes[name] == self.full_detector_response.axes[name]:
                print(f"Warning: the axis {name} is not consistent between the event and response!")
                return False

        print(f"    --> pass")
        return True

    def _modify_axes(self):
        """
        Modify the axes of data. This method will be removed in the future.
        """

        warnings.warn("Note that _modify_axes() in DataLoader was implemented for a temporary use. It will be removed in the future.", FutureWarning)
        warnings.warn("Make sure to perform _modify_axes() only once after the data are loaded.")

        if self.coordsys_conv_matrix.binning_method == 'Time':
            axis_name = ['Time', 'Em', 'Phi', 'PsiChi']

        elif self.coordsys_conv_matrix.binning_method == 'ScAtt':
            axis_name = ['ScAtt', 'Em', 'Phi', 'PsiChi']

        for name in axis_name:

            print(f"... checking the axis {name} of the event and background files...")
            
            event_edges, event_unit = self.event_dense.axes[name].edges, self.event_dense.axes[name].unit
            bkg_edges, bkg_unit = self.bkg_dense.axes[name].edges, self.bkg_dense.axes[name].unit

            if np.all(event_edges == bkg_edges):
                print(f"    --> pass (edges)") 
            else:
                print(f"Warning: the edges of the axis {name} are not consistent between the event and background!")
                print(f"        event      : {event_edges}")
                print(f"        background : {bkg_edges}")
                return False

            if event_unit == bkg_unit:
                print(f"    --> pass (unit)") 
            else:
                print(f"Warning: the unit of the axis {name} are not consistent between the event and background!")
                print(f"        event      : {event_unit}")
                print(f"        background : {bkg_unit}")
                return False

        # check the axes of the event/response files. 
        # Note that currently (2023-08-29) no unit is stored in the binned data. So only the edges are compared. This should be modified in the future.

        axis_name = ['Em', 'Phi', 'PsiChi']
        
        for name in axis_name:

            print(f"...checking the axis {name} of the event and response files...")

            event_edges, event_unit = self.event_dense.axes[name].edges, self.event_dense.axes[name].unit
            response_edges, response_unit = self.full_detector_response.axes[name].edges, self.full_detector_response.axes[name].unit
            
            if type(response_edges) == u.quantity.Quantity and self.is_miniDC2_format == True:
                response_edges = response_edges.value

            if np.all(event_edges == response_edges):
                print(f"    --> pass (edges)") 
            else:
                print(f"Warning: the edges of the axis {name} are not consistent between the event and background!")
                print(f"        event      : {event_edges}")
                print(f"        response : {response_edges}")
                return False

        axes_cds = Axes([self.event_dense.axes[0], \
                         self.full_detector_response.axes["Em"], \
                         self.full_detector_response.axes["Phi"], \
                         self.full_detector_response.axes["PsiChi"]])
        
        self.event_dense = Histogram(axes_cds, unit = self.event_dense.unit, contents = self.event_dense.contents)

        self.bkg_dense = Histogram(axes_cds, unit = self.bkg_dense.unit, contents = self.bkg_dense.contents)

        print(f"The axes in the event and background files are redefined. Now they are consistent with those of the response file.")

    '''
    def _check_sc_orientation_coverage(self):

        init_time_orientation = self.orientation.get_time()[0]
        init_time_event = Time(self.event_dense.axes["Time"].edges[0], format = 'unix')

        if not init_time_orientation <= init_time_event:
            print(f"Warning: the orientation file does not cover the observation")
            print(f"         initial time of the orientation file = {init_time_orientation}")
            print(f"         initial time of the event file       = {init_time_event}")
            return False

        end_time_orientation = self.orientation.get_time()[-1]
        end_time_event = Time(self.event_dense.axes["Time"].edges[-1], format = 'unix')

        if not end_time_event <= end_time_orientation:
            print(f"Warning: the orientation file does not cover the observation")
            print(f"         the end time of the orientation file = {end_time_orientation}")
            print(f"         the end time of the event file       = {end_time_event}")
            return False

        return True
    '''

    def load_full_detector_response_on_memory(self):
        """
        Load a response file on the computer memory.
        """

        axes_image_response = [self.full_detector_response.axes["NuLambda"], self.full_detector_response.axes["Ei"],
                               self.full_detector_response.axes["Em"], self.full_detector_response.axes["Phi"], self.full_detector_response.axes["PsiChi"]]

        self.image_response_dense = Histogram(axes_image_response, unit = self.full_detector_response.unit)

        nside = self.full_detector_response.axes["NuLambda"].nside
        npix = self.full_detector_response.axes["NuLambda"].npix 
    
        if self.is_miniDC2_format:
            for ipix in tqdm(range(npix)):
                self.image_response_dense[ipix] = np.sum(self.full_detector_response[ipix].to_dense(), axis = (4,5)) #Ei, Em, Phi, ChiPsi
        else:
            contents = self.full_detector_response._file['DRM']['CONTENTS'][:]
            self.image_response_dense[:] = contents * self.full_detector_response.unit

        self.response_on_memory = True

    ''' 
    def calc_coordsys_conv_matrix(self): 

        if not self._check_file_registration():
            print("Please load all files!")
            return 
    
        if not self._check_axis_consistency():
            print("Please the axes of the input files!")
            return 

#        if not self._check_sc_orientation_coverage():
#            print("Please the axes of the input files!")
#            return 

        print("... (DataLoader) calculating a coordinate conversion matrix...")
        
        # make an empty histogram for the response calculation
        axis_model_map = HealpixAxis(nside = self.full_detector_response.axes["NuLambda"].nside, 
                                     coordsys = "galactic", label = "lb")

        axis_coordsys_conv_matrix = [ axis_model_map, self.event_dense.axes["Time"], self.full_detector_response.axes["NuLambda"] ] #lb, Time, NuLambda

        self.coordsys_conv_matrix = Histogram(axis_coordsys_conv_matrix, unit = u.s, sparse = True)

        # calculate a dwell time map at each time bin and sky location
        nside = self.full_detector_response.axes["NuLambda"].nside
        npix = self.full_detector_response.axes["NuLambda"].npix 

        for ipix in tqdm(range(npix)):
            theta, phi = hp.pix2ang(nside, ipix)
            l, b = phi, np.pi/2 - theta

            pixel_coord = SkyCoord(l, b, unit = u.rad, frame = 'galactic')

            for i_time, [init_time, end_time] in enumerate(self.coordsys_conv_matrix.axes["Time"].bounds):
                init_time = Time(init_time, format = 'unix')
                end_time = Time(end_time, format = 'unix')
    
                filtered_orientation = self.orientation.source_interval(init_time, end_time)
                pixel_movement = filtered_orientation.get_target_in_sc_frame(target_name = f"pixel_{ipix}_{i_time}",
                                                                             target_coord = pixel_coord,
                                                                             quiet = True)

                time_diff = filtered_orientation.get_time_delta()

                dwell_time_map = filtered_orientation.get_dwell_map(response = self.full_detector_response.filename.resolve(),
                                                                    dts = time_diff,
                                                                    src_path = pixel_movement,
                                                                    quiet = True)

                self.coordsys_conv_matrix[ipix,i_time] = dwell_time_map.data * dwell_time_map.unit
                # (HealpixMap).data returns the numpy array without its unit.

        self.calc_image_response_projected()

    def save_coordsys_conv_matrix(self, filename = "coordsys_conv_matrix.hdf5"): 
        self.coordsys_conv_matrix.write(filename, overwrite = True)

    def load_coordsys_conv_matrix_from_filepath(self, filepath):

        if not self._check_file_registration():
            print("Please load all files!")
            return 
    
        if not self._check_axis_consistency():
            print("Please the axes of the input files!")
            return 

#        if not self._check_sc_orientation_coverage():
#            print("Please the axes of the input files!")
#            return 

        print("... (DataLoader) loading a coordinate conversion matrix...")

        self.coordsys_conv_matrix = Histogram.open(filepath)

        if not self.coordsys_conv_matrix.is_sparse:
            self.coordsys_conv_matrix = self.coordsys_conv_matrix.to_sparse()

        print(f"... checking the axes of the coordinate conversion matrix ...")

        if self.coordsys_conv_matrix.unit == u.s:
            print(f"    --> pass (unit)")
        else:
            print(f"Warning: the unit is wrong {self.coordsys_conv_matrix.unit}")
            return False

        axis_model_map = HealpixAxis(nside = self.full_detector_response.axes["NuLambda"].nside, 
                                     coordsys = "galactic", label = "lb")

        if self.coordsys_conv_matrix.axes['lb'] == axis_model_map:
            print(f"    --> pass (axis lb)")
        else:
            print(f"Warning: the axis of lb is inconsistent")
            return False

        if self.coordsys_conv_matrix.axes['Time'] == self.event_dense.axes["Time"]:
            print(f"    --> pass (axis Time)")
        else:
            print(f"Warning: the axis of Time is inconsistent")
            return False

        if self.coordsys_conv_matrix.axes['NuLambda'] == self.full_detector_response.axes['NuLambda']:
            print(f"    --> pass (axis NuLambda)")
        else:
            print(f"Warning: the axis of NuLambda is inconsistent")
            return False

        self.calc_image_response_projected()
    '''

    def calc_image_response_projected(self):
        """
        Calculate image_response_dense_projected, which is an intermidiate matrix used in RL algorithm.
        """

        print("... (DataLoader) calculating a projected image response ...")

        self.image_response_dense_projected = Histogram([ self.coordsys_conv_matrix.axes["lb"], self.full_detector_response.axes["Ei"] ],
                                                        unit = self.full_detector_response.unit * self.coordsys_conv_matrix.unit)

        if self.response_on_memory:

            self.image_response_dense_projected[:] = np.tensordot( np.sum(self.coordsys_conv_matrix, axis = (0)), 
                                                                   np.sum(self.image_response_dense, axis = (2,3,4)),
                                                                   axes = ([1], [0]) ) * self.full_detector_response.unit * self.coordsys_conv_matrix.unit
            # [Time/ScAtt, lb, NuLambda] -> [lb, NuLambda]
            # [NuLambda, Ei, Em, Phi, PsiChi] -> [NuLambda, Ei]
            # [lb, NuLambda] x [NuLambda, Ei] -> [lb, Ei]

        else:
            npix = self.full_detector_response.axes["NuLambda"].npix 

            for ipix in tqdm(range(npix)):
                if self.is_miniDC2_format:
                    full_detector_response_projected_Ei = np.sum(self.full_detector_response[ipix].to_dense(), axis = (1,2,3,4,5)) #Ei
                    # when np.sum is applied to a dense histogram, the unit is restored. when it is a sparse histogram, the unit is not restored. 
                else:
                    full_detector_response_projected_Ei = np.sum(self.full_detector_response[ipix].to_dense(), axis = (1,2,3)) #Ei
    
                coordsys_conv_matrix_projected_lb = np.sum(self.coordsys_conv_matrix[:,:,ipix], axis = (0)).todense() * self.coordsys_conv_matrix.unit #lb
    
                self.image_response_dense_projected += np.outer(coordsys_conv_matrix_projected_lb, full_detector_response_projected_Ei)



# Import
from cosipy.data_io import UnBinnedData 
import matplotlib.pyplot as plt
import numpy as np
import sys
import pandas as pd

try:
    # Load MEGAlib into ROOT
    import ROOT as M
    M.gSystem.Load("$(MEGAlib)/lib/libMEGAlib.so")

    # Initialize MEGAlib
    G = M.MGlobal()
    G.Initialize()
    
except:
    pass

class ReadTraTest(UnBinnedData):

    """Old method for reading tra file, used for unit testing."""
    
    def read_tra_old(self,make_plots=True):
        
        """Reads in MEGAlib .tra (or .tra.gz) file.
       
        Extended Summary
        ----------------
        This method uses MEGAlib to read events from the tra file. 
        This is used to compare to the new event reader, which is 
        independent of MEGAlib. 
        

        Parameters
        ----------
        make_plots : bool, optional
            Option to make binning plot.

        Returns
        -------
        cosi_dataset : dict
            Returns COSI dataset as a dictionary of the form:
            cosi_dataset = {'Full filename':self.data_file,
                        'Energies':erg,
                        'TimeTags':tt,
                        'Xpointings':np.array([lonX,latX]).T,
                        'Ypointings':np.array([lonY,latY]).T,
                        'Zpointings':np.array([lonZ,latZ]).T,
                        'Phi':phi,
                        'Chi local':chi_loc,
                        'Psi local':psi_loc,
                        'Distance':dist,
                        'Chi galactic':chi_gal,
                        'Psi galactic':psi_gal}
        
        Note
        ----
        This method sets the instance attribute, cosi_dataset, 
        but it does not explicitly return this.  
        """

        # tra file to use:
        tra_file = self.data_file

        # Make print statement:
        print()
        print("Read tra test...")
        print()
         
        # Check if file exists:
        Reader = M.MFileEventsTra()
        if Reader.Open(M.MString(tra_file)) == False:
            print("Unable to open file %s. Aborting!" %self.data_file)
            sys.exit()

        # Initialise empty lists:
            
        # Total photon energy
        erg = []
        # Time tag in UNIX time
        tt = []
        # Event Type (0: CE; 4:PE; ...)
        et = []
        # Latitude of X direction of spacecraft
        latX = []
        # Lontitude of X direction of spacecraft
        lonX = []
        # Latitude of Z direction of spacecraft
        latZ = []
        # Longitude of Z direction of spacecraft
        lonZ = []
        # Compton scattering angle
        phi = []
        # Measured data space angle chi (azimuth direction; 0..360 deg)
        chi_loc = []
        # Measured data space angle psi (polar direction; 0..180 deg)
        psi_loc = []
        # First lever arm distance in cm
        dist = []
        # Measured gal angle chi (lon direction)
        chi_gal = []
        # Measured gal angle psi (lat direction)
        psi_gal = [] 

        # Browse through tra file, select events, and sort into corresponding list:
        # Note: The Reader class from MEGAlib knows where an event starts and ends and
        # returns the Event object which includes all information of an event.
        # Note: Here only select Compton events (will add Photo events later as optional).
        # Note: All calculations and definitions taken from:
        # /MEGAlib/src/response/src/MResponseImagingBinnedMode.cxx.

        while True:
            
            Event = Reader.GetNextEvent()
            if not Event:
                break
                
            # Total Energy:
            erg.append(Event.Ei())
            # Time tag in UNIX seconds:
            tt.append(Event.GetTime().GetAsSeconds())
            # Event type (0 = Compton, 4 = Photo):
            et.append(Event.GetEventType())
            # x axis of space craft pointing at GAL latitude:
            latX.append(Event.GetGalacticPointingXAxisLatitude())
            # x axis of space craft pointing at GAL longitude:
            lonX.append(Event.GetGalacticPointingXAxisLongitude())
            # z axis of space craft pointing at GAL latitude:
            latZ.append(Event.GetGalacticPointingZAxisLatitude())
            # z axis of space craft pointing at GAL longitude:
            lonZ.append(Event.GetGalacticPointingZAxisLongitude())    
            # Compton scattering angle:
            phi.append(Event.Phi()) 
            # Data space angle chi (azimuth):
            chi_loc.append((-Event.Dg()).Phi())
            # Data space angle psi (polar):
            psi_loc.append((-Event.Dg()).Theta())
            # Interaction length between first and second scatter in cm:
            dist.append(Event.FirstLeverArm())
            # Gal longitude angle corresponding to chi:
            chi_gal.append((Event.GetGalacticPointingRotationMatrix()*Event.Dg()).Phi())
            # Gal longitude angle corresponding to chi:
            psi_gal.append((Event.GetGalacticPointingRotationMatrix()*Event.Dg()).Theta())
                
        # Initialize arrays:
        erg = np.array(erg)
        tt = np.array(tt)
        et = np.array(et)

        latX = np.array(latX)
        lonX = np.array(lonX)
        # Change longitudes to from 0..360 deg to -180..180 deg
        lonX[lonX > np.pi] -= 2*np.pi

        latZ = np.array(latZ)
        lonZ = np.array(lonZ)
        # Change longitudes to from 0..360 deg to -180..180 deg
        lonZ[lonZ > np.pi] -= 2*np.pi

        phi = np.array(phi)

        chi_loc = np.array(chi_loc)
        self.chi_loc_old = chi_loc

        # Change azimuth angle to 0..360 deg
        chi_loc[chi_loc < 0] += 2*np.pi

        psi_loc = np.array(psi_loc)
        self.psi_loc_old = psi_loc
        
        # For comparing chi_loc, psi_loc=0 values are arbitrary,
        # so we exclude them from the comparison. 
        psi_zero_index = psi_loc == 0

        dist = np.array(dist)

        chi_gal = np.array(chi_gal)
        psi_gal = np.array(psi_gal)
        self.chi_gal_old = chi_gal
        self.psi_gal_old = psi_gal
         
        # Construct Y direction from X and Z direction
        lonlatY = self.construct_scy(np.rad2deg(lonX),np.rad2deg(latX),
                                np.rad2deg(lonZ),np.rad2deg(latZ))
        lonY = np.deg2rad(lonlatY[0])
        latY = np.deg2rad(lonlatY[1])

        # Avoid negative zeros
        chi_loc[np.where(chi_loc == 0.0)] = np.abs(chi_loc[np.where(chi_loc == 0.0)])
       
        # Make observation dictionary
        cosi_dataset = {'Energies':erg,
                        'TimeTags':tt,
                        'Xpointings':np.array([lonX,latX]).T,
                        'Ypointings':np.array([lonY,latY]).T,
                        'Zpointings':np.array([lonZ,latZ]).T,
                        'Phi':phi,
                        'Chi local':self.chi_loc_old,
                        'Psi local':self.psi_loc_old,
                        'Distance':dist,
                        'Chi galactic':self.chi_gal_old,
                        'Psi galactic':self.psi_gal_old} 
        self.cosi_dataset = cosi_dataset

        # Write unbinned data to file (either fits or hdf5):
        self.write_unbinned_output() 
        
        return 




# Imports:
import matplotlib.pyplot as plt

class MakePlots:

    def make_basic_plot(self, x, y, plt_scale='loglog', savefig='plot.pdf',\
        x_error=[], plot_kwargs={}, fig_kwargs={}):

        """
        Make a basic plot.

        Inputs:
        x [list or array]: x-axis data to be plotted
        y [list or array]: y-axis data to be plotted

        Optional:
        plt_scale: scale of axes: loglog, semilogx, or semilogy.
        x_error [list or array]: x error bars
        savefig [str]: Name of saved figure
        plot_kwargs [dict]: pass any kwargs to plt.plot()
        fig_kwargs [dict]: pass any kwargs to plt.gca().set()
        """
        
        # Setup figure:
        ax = plt.gca()

        # Main plot:
        if plt_scale == "loglog":
            plt.loglog(x, y, **plot_kwargs)
        if plt_scale == "semilogx":
            plt.semilogx(x, y, **plot_kwargs)
        if plt_scale == "semilogy":
            plt.semilogy(x, y, **plot_kwargs)

        # Include x error bars:
        if len(x_error) != 0:
            
            # Remove label if defined:
            if "label" in plot_kwargs.keys():
                plot_kwargs["label"] = "_nolabel_"

            plt.errorbar(x, y, xerr=x_error, **plot_kwargs)

        # axes and labels:
        plt.grid(color="grey",alpha=0.3,ls="--")
        ax.set(**fig_kwargs)
        if "label" in plot_kwargs.keys():
            plt.legend(loc=1,frameon=True)
        
        # Save and show:
        plt.savefig(savefig)
        plt.show()
        plt.close()

        return



from .data_loader import DataLoader
from .modelmap import ModelMap
from .image_deconvolution import ImageDeconvolution
from .exposure_table import SpacecraftAttitudeExposureTable
from .coordsys_conversion_matrix import CoordsysConversionMatrix
